import torch


def chi_squared(
    estimate: torch.Tensor,
    target: torch.Tensor,
    uncertainty: torch.Tensor | float
) -> torch.Tensor:
    return torch.sum((estimate - target)**2 / uncertainty**2)


from ase import Atoms
import numpy as np

def generate_random_closest_approach(
    element_counts: dict[str, int],
    number_density: float,
    closest_approach: float = 0.0,
    max_steps_per_particle: int = 10_000
) -> Atoms:
    total_atoms = sum(element_counts.values())    

    # Get cell
    volume = total_atoms / number_density
    cell_length = volume**(1/3)
    cell: np.ndarray = cell_length * np.eye(3)
    inverse_cell = np.linalg.inv(cell)

    # Generate positions
    positions = np.zeros((total_atoms, 3))
    
    for particle in range(1, total_atoms):
        for attempt in range(max_steps_per_particle):
            proposed_position = np.random.rand(3) @ cell

            # Get distances with MIC
            displacements = proposed_position[None, :] - positions[:particle]
            displacement_shifts = np.round(displacements @ inverse_cell) @ cell
            image_displacements = displacements - displacement_shifts

            distances = np.linalg.norm(image_displacements, axis=-1)

            if np.min(distances) > closest_approach:
                positions[particle] = proposed_position
                break
            elif attempt == max_steps_per_particle - 1:
                raise RuntimeError(f"Unable to place atoms within max_steps_per_particle... Got to {particle+1}.")

    symbols = np.concatenate(list([element] * count for element, count in element_counts.items()))

    return Atoms(symbols, positions, cell=cell, pbc=True)


"""
Contains methods for calculating various spectra from a structure, and converting between
different types of spectra.
"""

from itertools import combinations_with_replacement

import torch

from ._data import NEUTRON_SCATTERING_LENGTHS, XRAY_FORM_FACTOR_PARAMS
import _cuda_utils

_DEFAULT_KERNEL_WIDTH = 0.035

def get_distance_matrix(
    points_1: torch.Tensor,
    points_2: torch.Tensor,
    cell: torch.Tensor
) -> torch.Tensor:
    """Gets distance matrix while accounting for minimal image convention.

    Args:
    - `points_1`: Tensor with shape M x D.
    - `points_2`: Tensor with shape N x D.
    - `cell`: D x D tensor of supercell vectors, e.g., [[ a1... ], [ a2... ], [ a3... ]].

    Returns: M x N tensor of distances.
    """
    pair_differences = points_2[None, :, :] - points_1[:, None, :]

    pair_differences_shifts = torch.round(pair_differences @ cell.inverse()) @ cell
    image_pair_differences = pair_differences - pair_differences_shifts

    distance_matrix = torch.linalg.norm(image_pair_differences, dim=-1)

    return distance_matrix


def get_observed_distances(
    points_1: torch.Tensor,
    points_2: torch.Tensor,
    cell: torch.Tensor
) -> torch.Tensor:
    """Gets the list of observed distances between two sets of points in D dimensions.
    
    If the points are identical, symmetry and zeros are ignored.

    Args:
    - `points_1`: Tensor with shape M x D
    - `points_2`: Tensor with shape N x D
    - `cell`: D x D tensor of supercell vectors, e.g., [[ a1... ], [ a2... ], [ a3... ]].
    
    Returns: 1D tensor of observed distances.
    """
    distance_matrix = get_distance_matrix(points_1, points_2, cell)

    if points_1 is points_2:
        # Ignore diagonal and symmetry-equivalent points
        mask = torch.triu(
            torch.ones_like(distance_matrix, dtype=torch.bool),
            diagonal=1
        )
        observed_distances = distance_matrix[mask]
    else:
        observed_distances = torch.flatten(distance_matrix)

    return observed_distances


def get_cell_volume(cell: torch.Tensor) -> torch.Tensor:
    """Calculates the volume of a super- or unit cell by its vectors.

    Note: The determinent may not be implemented on MPS, so the computation may briefly be moved
    to the CPU.

    Args:
    - `cell`: A D x D tensor of supercell vectors, e.g., [[ a1... ], [ a2... ], [ a3... ]].

    Returns: The volume of the cell as a tensor.
    """
    try:
        return cell.det().abs()
    except NotImplementedError:
        # For MPS
        return cell.cpu().det().abs().to(cell.device)


def get_partial_rdf(
    points_1: torch.Tensor,
    points_2: torch.Tensor,
    cell: torch.Tensor,
    bins: torch.Tensor,
    kernel_width: float = _DEFAULT_KERNEL_WIDTH,
) -> torch.Tensor:
    """Gets the partial radial distribution function (RDF) between two sets of points.

    Note: Uses kernel density estimation (KDE) with Gaussians so that the output is differentiable.

    Args:
    - `points_1`: Tensor with shape M x D.
    - `points_2`: Tensor with shape N x D.
    - `cell`: D x D tensor of supercell vectors, e.g., [[ a1... ], [ a2... ], [ a3... ]].
    - `bins`: The values of $r$ on which to evaluate the partial RDF.
    - `kernel_width`: Width of Gaussians for KDE. You should use a value less than half the width
    of the smallest expected peak. Using a good value may require trial and error.
    - `chunks`: In the event that there is not enough memory to compute all Gaussians before summing,
    the operation will be caried out in this many chunks.
    
    Returns: 1D tensor of estimated partial RDF values evaluated on `bins`.
    """
    observed_distances = get_observed_distances(points_1, points_2, cell)

    # Filter distances to ignore those that won't contribute to the RDF
    min_bin = torch.min(bins) - 3 * kernel_width
    max_bin = torch.max(bins) + 3 * kernel_width
    relevant_distances = observed_distances[
        torch.logical_and(
            min_bin < observed_distances,
            observed_distances < max_bin
        )
    ]

    # Calculate sum of gaussians
    try:
        gaussians: torch.Tensor = torch.exp(
            -0.5 * ((bins[None, :] - relevant_distances[:, None]) / kernel_width)**2
        ) / (kernel_width * (2 * torch.pi)**0.5)

        summed_gaussians = gaussians.sum(dim=0)
    except (torch.cuda.OutOfMemoryError):
        summed_gaussians = torch.zeros_like(bins)

        for device, distances_chunk in _cuda_utils.split_chunk(relevant_distances):
            bins = bins.to(device)
            summed_gaussians = summed_gaussians.to(device)

            gaussian_chunks: torch.Tensor = torch.exp(
                -0.5 * ((bins[None, :] - distances_chunk[:, None]) / kernel_width)**2
            ) / (kernel_width * (2 * torch.pi)**0.5)

            summed_gaussians += gaussian_chunks.sum(dim=0)

    # Scale to get RDF
    volume = get_cell_volume(cell)
    partial_rdf = (volume / (observed_distances).size(0)) * (1 / (4 * torch.pi * bins**2)) * summed_gaussians

    return partial_rdf


def get_neutron_total_rdf(
    chemical_symbols: list[str],
    positions: torch.Tensor,
    cell: torch.Tensor,
    bins: torch.Tensor,
    kernel_width: float = _DEFAULT_KERNEL_WIDTH,
    scattering_lengths: dict[str, float] = NEUTRON_SCATTERING_LENGTHS,
) -> torch.Tensor:
    """Calculates the total radial distribution function (RDF), G(r), for a structure.
    
    G(r) is defined as in "Keen. J. Appl. Crystallogr. (2001). 34, 172-177."

    Outputs in units of barns when scattering_lengths are in femtometers.

    Note: Uses kernel density estimation (KDE) with Gaussians so that the output is differentiable.

    Args:
    - `chemical_symbols`: List of atom symbols.
    - `positions: N x D tensor of atom positions in your system.
    - `cell`: D x D tensor of supercell vectors, e.g., [[ a1... ], [ a2... ], [ a3... ]].
    - `bins`: The values of $r$ on which to evaluate the total RDF.
    - `kernel_width`: Width of Gaussians for KDE. You should use a value less than half the width
    of the smallest expected peak in each partial RDF. Using a good value may require trial and
    error. Defaults to 0.05.
    - `scattering_lengths`: Dictionary that maps element labels to the element's average coherent
    bound neutron scattering length. Use units of femtometers. Defaults to values from _data.py file.

    Returns: 1D tensor of estimated total RDF values evaluated on `bins`.
    """
    total_particle_count = len(chemical_symbols)
    unique_elements = set(chemical_symbols)

    element_fractions = {
        element: chemical_symbols.count(element) / total_particle_count for element in unique_elements
    }
    element_masks = {
        element: [symbol == element for symbol in chemical_symbols] for element in unique_elements
    }    

    total_rdf = torch.zeros_like(bins)
    #for (element_1, element_2) in product(unique_elements, repeat=2):
    for element_1, element_2 in combinations_with_replacement(unique_elements, 2):
        positions_1 = positions[element_masks[element_1]]
        fraction_1 = element_fractions[element_1]
        scattering_length_1 = scattering_lengths[element_1]

        if element_1 != element_2:
            positions_2 = positions[element_masks[element_2]]
            fraction_2 = element_fractions[element_2]
            scattering_length_2 = scattering_lengths[element_2] 
        else:
            positions_2 = positions_1
            fraction_2 = fraction_1
            scattering_length_2 = scattering_length_1

        coefficient = fraction_1 * fraction_2 * scattering_length_1 * scattering_length_2 * 0.01

        if element_1 != element_2:
            coefficient *= 2

        partial_rdf = get_partial_rdf(
            positions_1,
            positions_2,
            cell,
            bins,
            kernel_width
        )

        total_rdf += coefficient * (partial_rdf - 1)

    return total_rdf


def get_neutron_total_correlation_function(
    total_rdf: torch.Tensor,
    r_bins: torch.Tensor,
    chemical_symbols: list[str],
    cell: torch.Tensor,
    scattering_lengths: dict[str, float] = NEUTRON_SCATTERING_LENGTHS,
):
    """Calculates the total correlation function, T(r), from the total RDF.

    T(r) is defined as in "Keen. J. Appl. Crystallogr. (2001). 34, 172-177."

    You can get the `total_rdf` from `get_neutron_total_rdf(...)`.

    Args:
    - `total_rdf`: 1D tensor of total radial distribution function values evaluated on `r_bins`.
    - `r_bins`: The values of $r$ on which the total_rdf is evaluated.
    - `chemical_symbols`: List of atom symbols.
    - `cell`: D x D tensor of supercell vectors, e.g., [[ a1... ], [ a2... ], [ a3... ]].
    - `scattering_lengths`: Dictionary that maps element labels to the element's average coherent
    bound neutron scattering length. Use units of femtometers. Defaults to values from _data.py file.

    Returns: 1D tensor of the total correlation function, T(r), evaluated on r_bins.
    """
    total_particle_count = len(chemical_symbols)
    unique_elements = set(chemical_symbols)

    element_fractions = {
        element: chemical_symbols.count(element) / total_particle_count for element in unique_elements
    }

    volume = get_cell_volume(cell)
    total_density = total_particle_count / volume

    rdf_offset = 0.01 * (sum(fraction * scattering_lengths[element] for element, fraction in element_fractions.items()))**2

    return 4 * torch.pi * r_bins * total_density * (total_rdf + rdf_offset)
    

def get_neutron_total_scattering_sf(
    total_rdf: torch.Tensor,
    r_bins: torch.Tensor,
    q_bins: torch.Tensor,
    chemical_symbols: list[str],
    cell: torch.Tensor,
) -> torch.Tensor:
    """Calculates the total scattering structure factor (SF), F(Q) = i(Q), from the total RDF.

    F(Q) = i(Q) is defined as in "Keen. J. Appl. Crystallogr. (2001). 34, 172-177."

    You can get the `total_rdf` from `get_neutron_total_rdf(...)`.

    Note: It is important for the accuracy of this calculation that `r_bins` is large enough that
    the `total_rdf` becomes close to zero.

    Args:
    - `total_rdf`: 1D tensor of total radial distribution function values evaluated on `r_bins`.
    - `r_bins`: The values of $r$ on which the total_rdf is evaluated.
    - `q_bins`: The values of $Q$ on which to evaluate the total scattering SF.
    - `chemical_symbols`: List of atom symbols.
    - `cell`: D x D tensor of supercell vectors, e.g., [[ a1... ], [ a2... ], [ a3... ]].

    Returns: 1D tensor of estimated total-scattering SF values, evaluated on q_bins.
    """
    total_particle_count = len(chemical_symbols)
    volume = get_cell_volume(cell)
    total_density = total_particle_count / volume

    integrand = r_bins[None, :] * total_rdf[None, :] * torch.sin(q_bins[:, None] * r_bins[None, :]) / q_bins[:, None]

    total_scattering_sf = 4 * torch.pi * total_density * torch.trapz(integrand, r_bins)
    return total_scattering_sf

def get_xray_total_scattering_sf(
    chemical_symbols: list[str],
    positions: torch.Tensor,
    cell: torch.Tensor,
    q_bins: torch.Tensor,
    r_bins: torch.Tensor,
    kernel_width: float = _DEFAULT_KERNEL_WIDTH,
    form_factor_params: dict[str, dict[str, list[float]]] = XRAY_FORM_FACTOR_PARAMS,
) -> torch.Tensor:
    """Calculates the total X-ray scattering factor (SF), F^X(r), for a structure.
    
    F^X(r) is defined as in "Keen. J. Appl. Crystallogr. (2001). 34, 172-177."

    Outputs in units of barns when scattering_lengths are in femtometers.

    Note: Uses kernel density estimation (KDE) with Gaussians so that the output is differentiable.

    Args:
    - `chemical_symbols`: List of atom symbols.
    - `positions: N x D tensor of atom positions in your system.
    - `cell`: D x D tensor of supercell vectors, e.g., [[ a1... ], [ a2... ], [ a3... ]].
    - `bins`: The values of $r$ on which to evaluate the total RDF.
    - `kernel_width`: Width of Gaussians for KDE. You should use a value less than half the width
    of the smallest expected peak in each partial RDF. Using a good value may require trial and
    error. Defaults to 0.05.
    - `scattering_lengths`: Dictionary that maps element labels to the element's average coherent
    bound neutron scattering length. Use units of femtometers. Defaults to values from _data.py file.

    Returns: 1D tensor of estimated total RDF values evaluated on `bins`.
    """
    total_particle_count = len(chemical_symbols)
    volume = get_cell_volume(cell)
    total_density = total_particle_count / volume

    unique_elements = set(chemical_symbols)

    element_fractions = {
        element: chemical_symbols.count(element) / total_particle_count for element in unique_elements
    }

    element_masks = {
        element: [symbol == element for symbol in chemical_symbols] for element in unique_elements
    }

    # Form factors
    element_form_factors: dict[str, torch.Tensor] = {}
    for element in unique_elements:
        a = torch.tensor(form_factor_params[element]["a"], device=q_bins.device)
        b = torch.tensor(form_factor_params[element]["b"], device=q_bins.device)
        c = form_factor_params[element]["c"][0]

        element_form_factors[element] = c + torch.sum(a[:, None] * torch.exp(-b[:, None] * (q_bins[None, :] / (4 * torch.pi))**2), dim=0)

    sum_of_form_factors = sum(element_fractions[element] * element_form_factors[element] for element in unique_elements)

    scaled_form_factors = {
        element: element_fractions[element] * element_form_factors[element] / sum_of_form_factors for element in unique_elements
    }

    qr_product = r_bins[:, None] * q_bins[None, :]
    total_sf = torch.zeros_like(q_bins)
    for element_1, element_2 in combinations_with_replacement(unique_elements, 2):
        positions_1 = positions[element_masks[element_1]]
        scaled_form_factor_1 = scaled_form_factors[element_1]

        if element_1 != element_2:
            positions_2 = positions[element_masks[element_2]]
            scaled_form_factor_2 = scaled_form_factors[element_2] 
        else:
            positions_2 = positions_1
            scaled_form_factor_2 = scaled_form_factor_1

        coefficient = total_density * scaled_form_factor_1 * scaled_form_factor_2

        if element_1 != element_2:
            coefficient *= 2

        partial_rdf = get_partial_rdf(
            positions_1,
            positions_2,
            cell,
            r_bins,
            kernel_width
        )

        integrand = 4 * torch.pi * r_bins[:, None]**2 * (partial_rdf[:, None] - 1) * torch.sin(qr_product) / qr_product

        total_sf += coefficient * torch.trapz(integrand, r_bins, dim=0)

    return total_sf


def get_xray_total_rdf(
    total_sf: torch.Tensor,
    r_bins: torch.Tensor,
    q_bins: torch.Tensor,
    chemical_symbols: list[str],
    cell: torch.Tensor,
) -> torch.Tensor:
    total_particle_count = len(chemical_symbols)
    volume = get_cell_volume(cell)
    total_density = total_particle_count / volume
    
    qr_product = q_bins[None, :] * r_bins[:, None]
    integrand = 4 * torch.pi * q_bins[None, :]**2 * total_sf[None, :] * torch.sin(qr_product) / qr_product

    total_rdf = torch.trapz(integrand, q_bins) / ((2 * torch.pi)**3 * total_density)

    return total_rdf


from functools import cache
from typing import Generator

import torch


@cache
def get_device_count():
    import nvidia_smi
    nvidia_smi.nvmlInit()

    return nvidia_smi.nvmlDeviceGetCount()
    

def split_chunk(tensor: torch.Tensor) -> Generator[tuple[str, torch.Tensor], None, None]:
    device_count = get_device_count()

    for device_idx, chunk in enumerate(torch.chunk(tensor, device_count)):
        device = f"cuda:{device_idx}"
        chunk = chunk.to(device)

        yield device, chunk



import torch


def get_coulomb_potential(
    observed_distances: torch.Tensor,
    charge_1: int,
    charge_2: int,
    relative_permittivity: float = 1.0,
) -> torch.Tensor:
    """Calculates the Coulomb potential between two pairs of particles.

    Args:
    - `observed_distances`: Tensor of observed distances, in Angstroms.
    - `charge_1`, `charge_2`: Charges of each pair of particles (in units of electron-charges)
    - `relative_permittivity` (Optional): The dieletric constant. The default value is 1.0.

    Returns: The Coulomb potential in units of eV, as a tensor.
    """
    # Prefactor is e^2 / (4 pi epsilon)
    prefactor = 14.39 / relative_permittivity # eV Angstrom

    contributions = prefactor * (charge_1 * charge_2) / observed_distances
    return torch.sum(contributions)


def get_lennard_jones_potential(
    observed_distances: torch.Tensor,
    sigma: float,
    epsilon: float,
) -> torch.Tensor:
    """Calculates the Lennard-Jones potential between two pairs of particles.

    V = 4 * epsilon * [(sigma / r)^12 - (sigma / r)^6].

    Note: r_min = 2^(1/6) sigma.

    Args:
    - `observed_distances`: Tensor of observed distances, in Angstroms.
    - `sigma`: The Lennard-Jones sigma parameter, in Angstroms.
    - `epsilon`: The Lennard-Jones epsilon parameter, in eV.

    Returns: The Lennard-Jones potential in units of eV, as a tensor.
    """
    term_6 = (sigma / observed_distances)**6
    term_12 = term_6**2

    contributions = 4 * epsilon * (term_12 - term_6)
    return torch.sum(contributions)



"""Holds useful data such as scattering lengths."""
import numpy as np

# Bound coherent scattering lengths, in fm, from https://www.ncnr.nist.gov/resources/n-lengths/
NEUTRON_SCATTERING_LENGTHS = {
    "Si": 4.1491,
    "O": 5.803,
}

# From http://lampx.tugraz.at/~hadley/ss1/crystaldiffraction/atomicformfactors/formfactors.php
XRAY_FORM_FACTOR_PARAMS = {
    "Si": {
        "a": [6.2915, 3.0353, 1.9891, 1.541],
        "b": [2.4386, 32.3337, 0.6785, 81.6937],
        "c": [1.1407]
    },
    "O": {
        "a": [3.0485, 2.2868, 1.5463, 0.867],
        "b": [13.2771, 5.7011, 0.3239, 32.9089],
        "c": [0.2508],
    }
}


from typing import Callable
import time

from ase import Atoms
from ase.calculators.calculator import Calculator, all_properties
import torch
import numpy as np

from .spectra import get_cell_volume


class GBSDCalculator(Calculator):
    implemented_properties = ['energy', 'free_energy', 'forces', 'stress']

    def __init__(
        self,
        energy_method: Callable[[torch.Tensor, list[str], torch.Tensor], torch.Tensor],
        device: torch.device | str | None = None,
        **kwargs
    ):
        super().__init__(**kwargs)

        self.energy_method = energy_method
        self.device = device

        self._time_spent_converting = 0.

    # def _move_to_device_and_time(self, array: np.ndarray) -> torch.Tensor:
    #     start_time = time.perf_counter()
    #     tensor = torch.from_numpy(array).to(self.device)
    #     self._time_spent_converting += time.perf_counter() - start_time

    #     return tensor
    
    # def _move_to_cpu_and_time(self, tensor: torch.Tensor) -> np.ndarray:
    #     start_time = time.perf_counter()
    #     array = tensor.numpy(force=True)
    #     self._time_spent_converting += time.perf_counter() - start_time

    #     return array

    def calculate(
        self,
        atoms: Atoms | None = None,
        properties: list[str] | None = None,
        system_changes: list[str] | None = None,
    ) -> None:
        super().calculate(atoms, properties, system_changes)

        properties = properties or all_properties

        if atoms is None:
            raise ValueError("No atoms.")
        
        fractional_positions = torch.from_numpy(atoms.get_scaled_positions()).to(self.device)
        species = atoms.get_chemical_symbols()
        cell = torch.from_numpy(atoms.get_cell().array).to(self.device)
        
        strain = torch.zeros_like(cell, requires_grad=True)
        cell = cell @ (torch.eye(3, device=self.device) + strain)

        positions = fractional_positions @ cell

        energy = self.energy_method(positions, species, cell)
        if "energy" in properties:
            self.results.update(energy=float(energy))

        if "free_energy" in properties:
            self.results.update(free_energy=float(energy))

        if "forces" in properties:
            forces = - torch.autograd.grad(energy, positions)[0]
            self.results.update(forces=forces.numpy(force=True))

        if "stress" in properties:
            stress = (1 / get_cell_volume(cell)) * torch.autograd.grad(energy, strain)[0]
            self.results.update(stress=stress.numpy(force=True).flat[[0, 4, 8, 5, 2, 1]])  # voigt stress
